# Copyright 2004 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

from twisted.web.microdom import Element
from twisted.protocols import msn
from legacy import glue
import debug



class SubscriptionManager:
	def __init__(self, session):
		self.session = session
		self.subscriptionBuffer = []
	
	def removeMe(self):
		self.session = None
	
	def syncJabberLegacyLists(self):
		""" Synchronises the MSN contact list on server with the Jabber contact list """
		# We have to make an MSNContactList from the XDB data, then compare it with the one the server sent
		# Any subscription changes must be sent to the client, as well as changed in the XDB
		debug.log("Subscriptions: Session \"%s\" starting syncJabberLegacyLists()" % (self.session.jabberID))
		result = self.session.pytrans.xdb.request(self.session.jabberID, "jabber:iq:roster")
		oldContactList = msn.MSNContactList()
		if(result):
			for item in result.childNodes:
				user = item.getAttribute("jid")
				sub = item.getAttribute("subscription")
				lists = glue.jabsub2msnlist(sub)
				contact = msn.MSNContact(userHandle=user, screenName="", lists=lists)
				oldContactList.addContact(contact)
		
		newXDB = Element("query")
		newXDB.namespace = "jabber:iq:roster"
		
		contactList = self.session.legacycon.getContacts()
		for contact in contactList.getContacts().values():
			# Compare with the XDB <item/> entry
			oldContact = oldContactList.getContact(contact.userHandle)
			if(oldContact == None):
				oldLists = 0
			else:
				oldLists = oldContact.lists
			lists = contact.lists
			
			
			def updatePresence(ptype): # Convienence
				self.session.sendPresence(to=self.session.jabberID, fro=glue.msn2jid(contact.userHandle), ptype=ptype)
			
			if(not (oldLists & msn.FORWARD_LIST)):
				if(lists & msn.FORWARD_LIST):
					# User has been added to forward list
					updatePresence("subscribed")
			else:
				if(not (lists & msn.FORWARD_LIST)):
					# User has been removed from forward list
					updatePresence("unsubscribed")
			
			if(not (oldLists & msn.REVERSE_LIST)):
				if(lists & msn.REVERSE_LIST):
					# User has been added to reverse list
					if(not (lists & msn.BLOCK_LIST)):
						# User isn't blocked, so will subscribe to your presence
						updatePresence("subscribe")
			else:
				if(not (lists & msn.REVERSE_LIST)):
					# User has been removed from reverse list
					updatePresence("unsubscribe")
			
			item = Element("item")
			item.setAttribute("jid", contact.userHandle)
			item.setAttribute("subscription", glue.msnlist2jabsub(lists))
			newXDB.appendChild(item)
		
		# Update the XDB
		self.session.pytrans.xdb.set(self.session.jabberID, "jabber:iq:roster", newXDB)
		debug.log("Subscriptions: Session \"%s\" finished syncJabberLegacyLists()" % (self.session.jabberID))
	
	
	
	def msnContactAddedMe(self, userHandle):
		# User has been added to our reverse list (we are now on their forward list)
		debug.log("Subscriptions: Session \"%s\" msnContactAddedMe(\"%s\")" % (self.session.jabberID, userHandle))
		
		# Update the XDB if needed
		document = self.session.pytrans.xdb.request(self.session.jabberID, "jabber:iq:roster")
		for item in document.childNodes:
			user = item.getAttribute("jid")
			if(user == userHandle):
				break
		else:
			# Not found, so we need to add it
			document = Element("item")
			item.setAttribute("jid", userHandle)
			item.setAttribute("subscription", "from")
			document.appendChild(item)
			self.session.pytrans.xdb.set(self.session.jabberID, "jabber:iq:roster", document)
		
		# Send a subscription packet
		source = glue.msn2jid(userHandle)
		self.session.sendPresence(to=self.session.jabberID, fro=source, ptype="subscribe")
	
	def msnContactRemovedMe(self, userHandle):
		# User has been removed from our reverse list (we are no longer on their forward list)
		debug.log("Subscriptions: Session \"%s\" msnContactRemovedMe(\"%s\")" % (self.session.jabberID, userHandle))
		
		# Update the XDB
		document = self.session.pytrans.xdb.request(self.session.jabberID, "jabber:iq:roster")
		for item in document.childNodes:
			user = item.getAttribute("jid")
			if(user == userHandle):
				document.removeChild(item)
				break
		self.session.pytrans.xdb.set(self.session.jabberID, "jabber:iq:roster", document)
		
		# Send an unsubscribe packet (the contact is no longer viewing the user's presence)
		source = glue.msn2jid(userHandle)
		self.session.sendPresence(to=self.session.jabberID, fro=source, ptype="unsubscribe")
	
	
	def flushSubscriptionBuffer(self):
		for (to, subtype) in self.subscriptionBuffer:
			self.jabberSubscriptionReceived(to, subtype)
	
	def jabberSubscriptionReceived(self, to, subtype):
		debug.log("Subscriptions: Session \"%s\" - jabberSubscriptionReceived(\"%s\", \"%s\")" % (self.session.jabberID, to, subtype))
		
		if(not self.session.ready):
			# Buffer until we have received the MSN contact list
			self.subscriptionBuffer.append((to, subtype))
			return
		
		def updatePresence(ptype): # Convienence
			self.session.sendPresence(to=self.session.jabberID, fro=to, ptype=ptype)
		
		if(to.find('@') > 0): # For contacts
			userHandle = glue.jid2msn(to)
			
			msnContact = self.session.legacycon.getContacts().getContact(userHandle)
			lists = 0 # Lists default to none
			if(msnContact and msnContact.lists):
				lists = msnContact.lists # We know extra information about what lists they're in.
			
			if(subtype == "subscribe"):
				# User wants to subscribe to contact's presence
				if(lists & msn.FORWARD_LIST):
					# The contact is already on the user's MSN list, so tell the user that this is so
					updatePresence("subscribed")
				else:
					# Add the contact to the user's MSN list
					def cb(arg=None):
						updatePresence("subscribed")
					self.session.legacycon.addContact(msn.FORWARD_LIST, userHandle).addCallback(cb)
			
			elif(subtype == "subscribed"):
				# The user has granted this contact subscription.
				def remFromBlockList(arg=None):
					if(lists & msn.BLOCK_LIST):
						# If they were on the block list remove them
						def cb(arg=None):
							pass
						self.session.legacycon.remContact(msn.BLOCK_LIST, userHandle).addCallback(cb)
				
				if(not (lists & msn.ALLOW_LIST)):
					# We add the contact to the allow list
					self.session.legacycon.addContact(msn.ALLOW_LIST, userHandle).addCallback(remFromBlockList)
			
			elif(subtype == "unsubscribe"):
				# User wants to unsubscribe to this contact's presence. (User is removing the contact from their list)
				if(lists & msn.FORWARD_LIST):
					# Contact is in the forward list, so remove them
					def cb(arg=None):
						pass
					self.session.legacycon.remContact(msn.FORWARD_LIST, userHandle).addCallback(cb)
				else:
					updatePresence("unsubscribed")
			
			elif(subtype == "unsubscribed"):
				# The user wants to remove this contact's authorisation. Contact will no longer be able to see user
				def addToBlockList(arg=None):
					if(not (lists & msn.BLOCK_LIST)):
						# If they're not on the block list then add them to it
						def cb(arg=None):
							pass
						self.session.legacycon.addContact(msn.BLOCK_LIST, userHandle).addCallback(cb)

				if(lists & msn.ALLOW_LIST):
					# If they're currently on the allow list then remove them
					# Should we add them to the block list?
					self.session.legacycon.remContact(msn.ALLOW_LIST, userHandle)#.addCallback(addToBlockList)
# 				else:
# 					addToBlockList()
		
		else: # The user wants to change subscription to the transport
			if(subtype == "subscribe"):
				updatePresence("subscribed")
			
			elif(subtype == "subscribed"):
				return # Nothing to do
			
			elif(subtype == "unsubscribe" or subtype == "unsubscribed"):
				# They want to unregister. Ok, we can do that
				jid = self.session.jabberID
				debug.log("Subscriptions: Session \"%s\" is about to be unregistered" % (jid))
				self.session.pytrans.registermanager.removeRegInfo(jid)
				debug.log("Subscriptions: Session \"%s\" has been unregistered" % (jid))
	